@@ -43,6 +43,7 @@
 
 #include "tracker-endpoint-dbus.h"
 #include "tracker-notifier.h"
+#include "tracker-notifier-private.h"
 #include "tracker-private.h"
 
 #include <gio/gio.h>
@@ -60,6 +61,13 @@ static const gchar introspection_xml[] =
 	"      <arg type='a{sv}' name='arguments' direction='in' />"
 	"      <arg type='as' name='result' direction='out' />"
 	"    </method>"
+	"    <method name='Serialize'>"
+	"      <arg type='s' name='query' direction='in' />"
+	"      <arg type='h' name='output_stream' direction='in' />"
+	"      <arg type='i' name='flags' direction='in' />"
+	"      <arg type='i' name='format' direction='in' />"
+	"      <arg type='a{sv}' name='arguments' direction='in' />"
+	"    </method>"
 	"    <method name='Update'>"
 	"      <arg type='h' name='input_stream' direction='in' />"
 	"    </method>"
@@ -70,6 +78,13 @@ static const gchar introspection_xml[] =
 	"      <arg type='h' name='input_stream' direction='in' />"
 	"      <arg type='aaa{ss}' name='result' direction='out' />"
 	"    </method>"
+	"    <method name='Deserialize'>"
+	"      <arg type='h' name='input_stream' direction='in' />"
+	"      <arg type='i' name='flags' direction='in' />"
+	"      <arg type='i' name='format' direction='in' />"
+	"      <arg type='s' name='default_graph' direction='in' />"
+	"      <arg type='a{sv}' name='arguments' direction='in' />"
+	"    </method>"
 	"    <signal name='GraphUpdated'>"
 	"      <arg type='sa{ii}' name='updates' />"
 	"    </signal>"
@@ -506,6 +521,84 @@ stmt_execute_cb (GObject      *object,
 	g_object_unref (task);
 }
 
+static void
+splice_rdf_cb (GObject      *object,
+               GAsyncResult *res,
+               gpointer      user_data)
+{
+	QueryRequest *request = user_data;
+	GError *error = NULL;
+
+	g_output_stream_splice_finish (G_OUTPUT_STREAM (object),
+	                               res, &error);
+
+	if (error) {
+		/* The query request method invocations has been already replied */
+		g_warning ("Error splicing RDF data: %s", error->message);
+		g_error_free (error);
+	}
+
+	query_request_free (request);
+}
+
+static void
+stmt_serialize_cb (GObject      *object,
+                   GAsyncResult *res,
+                   gpointer      user_data)
+{
+	QueryRequest *request = user_data;
+	GInputStream *istream;
+	GError *error = NULL;
+
+	istream = tracker_sparql_statement_serialize_finish (TRACKER_SPARQL_STATEMENT (object),
+	                                                     res, &error);
+	if (!istream) {
+		g_dbus_method_invocation_return_gerror (request->invocation, error);
+		g_error_free (error);
+		query_request_free (request);
+		return;
+	}
+
+	g_dbus_method_invocation_return_value (request->invocation, NULL);
+	g_output_stream_splice_async (G_OUTPUT_STREAM (request->data_stream),
+	                              istream,
+	                              G_OUTPUT_STREAM_SPLICE_CLOSE_SOURCE |
+	                              G_OUTPUT_STREAM_SPLICE_CLOSE_TARGET,
+	                              G_PRIORITY_DEFAULT,
+	                              request->global_cancellable,
+	                              splice_rdf_cb,
+	                              request);
+}
+
+static void
+serialize_cb (GObject      *object,
+              GAsyncResult *res,
+              gpointer      user_data)
+{
+	QueryRequest *request = user_data;
+	GInputStream *istream;
+	GError *error = NULL;
+
+	istream = tracker_sparql_connection_serialize_finish (TRACKER_SPARQL_CONNECTION (object),
+	                                                      res, &error);
+	if (!istream) {
+		g_dbus_method_invocation_return_gerror (request->invocation, error);
+		g_error_free (error);
+		query_request_free (request);
+		return;
+	}
+
+	g_dbus_method_invocation_return_value (request->invocation, NULL);
+	g_output_stream_splice_async (G_OUTPUT_STREAM (request->data_stream),
+	                              istream,
+	                              G_OUTPUT_STREAM_SPLICE_CLOSE_SOURCE |
+	                              G_OUTPUT_STREAM_SPLICE_CLOSE_TARGET,
+	                              G_PRIORITY_DEFAULT,
+	                              request->global_cancellable,
+	                              splice_rdf_cb,
+	                              request);
+}
+
 static void
 update_cb (GObject      *object,
            GAsyncResult *res,
@@ -602,6 +695,25 @@ read_update_blank_cb (GObject      *object,
 	                                              request);
 }
 
+static void
+deserialize_cb (GObject      *object,
+                GAsyncResult *res,
+                gpointer      user_data)
+{
+	UpdateRequest *request = user_data;
+	GError *error = NULL;
+
+	if (!tracker_sparql_connection_deserialize_finish (TRACKER_SPARQL_CONNECTION (object),
+	                                                   res, &error)) {
+		g_dbus_method_invocation_return_gerror (request->invocation, error);
+		update_request_free (request);
+		return;
+	}
+
+	g_dbus_method_invocation_return_value (request->invocation, NULL);
+	update_request_free (request);
+}
+
 static TrackerSparqlStatement *
 create_statement (TrackerSparqlConnection  *conn,
                   const gchar              *query,
@@ -721,6 +833,70 @@ endpoint_dbus_iface_method_call (GDBusConnection       *connection,
 		}
 
 		g_variant_iter_free (arguments);
+		g_free (query);
+	} else if (g_strcmp0 (method_name, "Serialize") == 0) {
+		TrackerSerializeFlags flags;
+		TrackerRdfFormat format;
+
+		if (tracker_endpoint_dbus_forbid_operation (endpoint_dbus,
+		                                            invocation,
+		                                            TRACKER_OPERATION_TYPE_SELECT)) {
+			g_dbus_method_invocation_return_error (invocation,
+			                                       G_DBUS_ERROR,
+			                                       G_DBUS_ERROR_ACCESS_DENIED,
+			                                       "Operation not allowed");
+			return;
+		}
+
+		g_variant_get (parameters, "(shiia{sv})", &query, &handle, &flags, &format, &arguments);
+
+		if (fd_list)
+			fd = g_unix_fd_list_get (fd_list, handle, &error);
+
+		if (fd < 0) {
+			g_dbus_method_invocation_return_error (invocation,
+			                                       G_DBUS_ERROR,
+			                                       G_DBUS_ERROR_INVALID_ARGS,
+			                                       "Did not get a file descriptor");
+		} else {
+			QueryRequest *request;
+
+			query = tracker_endpoint_dbus_add_prologue (endpoint_dbus,
+			                                            query);
+
+			request = query_request_new (endpoint_dbus, invocation, fd);
+
+			if (arguments) {
+				TrackerSparqlStatement *stmt;
+
+				stmt = create_statement (conn, query, arguments,
+				                         request->cancellable,
+				                         &error);
+				if (stmt) {
+					tracker_sparql_statement_serialize_async (stmt,
+					                                          flags,
+					                                          format,
+					                                          request->cancellable,
+					                                          stmt_serialize_cb,
+					                                          request);
+					/* Statements are single use here... */
+					g_object_unref (stmt);
+				} else {
+					query_request_free (request);
+					g_dbus_method_invocation_return_gerror (invocation,
+					                                        error);
+				}
+			} else {
+				tracker_sparql_connection_serialize_async (conn,
+				                                           flags,
+				                                           format,
+				                                           query,
+				                                           request->cancellable,
+				                                           serialize_cb,
+				                                           request);
+			}
+		}
+
 		g_free (query);
 	} else if (g_strcmp0 (method_name, "Update") == 0 ||
 	           g_strcmp0 (method_name, "UpdateArray") == 0) {
@@ -779,6 +955,49 @@ endpoint_dbus_iface_method_call (GDBusConnection       *connection,
 			request = update_request_new (endpoint_dbus, invocation, FALSE, fd);
 			update_request_read_next (request, read_update_blank_cb);
 		}
+	} else if (g_strcmp0 (method_name, "Deserialize") == 0) {
+		TrackerDeserializeFlags flags;
+		TrackerRdfFormat format;
+		gchar *graph;
+
+		if (tracker_endpoint_dbus_forbid_operation (endpoint_dbus,
+		                                            invocation,
+		                                            TRACKER_OPERATION_TYPE_UPDATE)) {
+			g_dbus_method_invocation_return_error (invocation,
+			                                       G_DBUS_ERROR,
+			                                       G_DBUS_ERROR_ACCESS_DENIED,
+			                                       "Operation not allowed");
+			return;
+		}
+
+		g_variant_get (parameters, "(hiisa{sv})", &handle, &flags, &format, &graph, &arguments);
+
+		if (fd_list)
+			fd = g_unix_fd_list_get (fd_list, handle, &error);
+
+		if (fd < 0) {
+			g_dbus_method_invocation_return_error (invocation,
+			                                       G_DBUS_ERROR,
+			                                       G_DBUS_ERROR_INVALID_ARGS,
+			                                       "Did not get a file descriptor");
+		} else {
+			TrackerSparqlConnection *conn;
+			UpdateRequest *request;
+
+			request = update_request_new (endpoint_dbus, invocation, FALSE, fd);
+			conn = tracker_endpoint_get_sparql_connection (TRACKER_ENDPOINT (request->endpoint));
+
+			tracker_sparql_connection_deserialize_async (conn,
+			                                             flags,
+			                                             format,
+			                                             graph && *graph ? graph : NULL,
+			                                             G_INPUT_STREAM (request->input_stream),
+			                                             request->endpoint->cancellable,
+			                                             deserialize_cb,
+			                                             request);
+		}
+
+		g_free (graph);
 	} else {
 		g_dbus_method_invocation_return_error (invocation,
 		                                       G_DBUS_ERROR,
@@ -859,6 +1078,7 @@ tracker_endpoint_dbus_initable_init (GInitable     *initable,
 
 	conn = tracker_endpoint_get_sparql_connection (endpoint);
 	endpoint_dbus->notifier = tracker_sparql_connection_create_notifier (conn);
+	tracker_notifier_disable_urn_query (endpoint_dbus->notifier);
 	g_signal_connect (endpoint_dbus->notifier, "events",
 	                  G_CALLBACK (notifier_events_cb), endpoint);
 
