@@ -19,11 +19,15 @@
 
 #include "config.h"
 
+/* Allow usage of tracker_namespace_manager_get_default() */
+#define TRACKER_VERSION_MIN_REQUIRED TRACKER_VERSION_3_2
+
 #include <glib.h>
 #include <json-glib/json-glib.h>
 
 #include <string.h>
 
+#include <tracker-deserializer-resource.h>
 #include <tracker-uri.h>
 #include <tracker-resource.h>
 #include <tracker-ontologies.h>
@@ -31,6 +35,9 @@
 /* For tracker_sparql_escape_string */
 #include "tracker-utils.h"
 
+/* For prefixed names parsing */
+#include <libtracker-sparql/core/tracker-sparql-grammar.h>
+
 #include <tracker-private.h>
 
 typedef struct {
@@ -39,7 +46,7 @@ typedef struct {
 	GHashTable *overwrite;
 } TrackerResourcePrivate;
 
-G_DEFINE_TYPE_WITH_PRIVATE (TrackerResource, tracker_resource, G_TYPE_OBJECT);
+G_DEFINE_TYPE_WITH_PRIVATE (TrackerResource, tracker_resource, G_TYPE_OBJECT)
 #define GET_PRIVATE(object)  (tracker_resource_get_instance_private (object))
 
 /**
@@ -49,10 +56,8 @@ G_DEFINE_TYPE_WITH_PRIVATE (TrackerResource, tracker_resource, G_TYPE_OBJECT);
  * @stability: Stable
  * @include: tracker-resource.h
  *
- * <para>
  * #TrackerResource keeps track of a set of properties for a given resource.
  * The resulting data can be serialized in several ways.
- * </para>
  */
 
 static char *
@@ -222,7 +227,7 @@ set_property (GObject      *object,
 
 /**
  * tracker_resource_new:
- * @identifier: A string containing a URI
+ * @identifier: (nullable): A string containing a URI, or %NULL.
  *
  * Creates a TrackerResource instance.
  *
@@ -284,7 +289,7 @@ tracker_resource_set_gvalue (TrackerResource *self,
 	g_hash_table_insert (priv->properties, g_strdup (property_uri), our_value);
 
 	g_hash_table_insert (priv->overwrite, g_strdup (property_uri), GINT_TO_POINTER (TRUE));
-};
+}
 
 static gboolean
 validate_boolean (gboolean    value,
@@ -350,7 +355,7 @@ validate_pointer (const void *pointer,
 		g_hash_table_insert (priv->overwrite,                               \
 		                     g_strdup (property_uri),                       \
 		                     GINT_TO_POINTER (TRUE));                       \
-	};
+	}
 
 /**
  * tracker_resource_set_boolean:
@@ -360,7 +365,7 @@ validate_pointer (const void *pointer,
  *
  * Sets a single-valued boolean object.
  */
-SET_PROPERTY_FOR_GTYPE (tracker_resource_set_boolean, gboolean, G_TYPE_BOOLEAN, g_value_set_boolean, validate_boolean);
+SET_PROPERTY_FOR_GTYPE (tracker_resource_set_boolean, gboolean, G_TYPE_BOOLEAN, g_value_set_boolean, validate_boolean)
 
 /**
  * tracker_resource_set_double:
@@ -370,7 +375,7 @@ SET_PROPERTY_FOR_GTYPE (tracker_resource_set_boolean, gboolean, G_TYPE_BOOLEAN,
  *
  * Sets a single-valued double object.
  */
-SET_PROPERTY_FOR_GTYPE (tracker_resource_set_double, double, G_TYPE_DOUBLE, g_value_set_double, validate_double);
+SET_PROPERTY_FOR_GTYPE (tracker_resource_set_double, double, G_TYPE_DOUBLE, g_value_set_double, validate_double)
 
 /**
  * tracker_resource_set_int:
@@ -380,7 +385,7 @@ SET_PROPERTY_FOR_GTYPE (tracker_resource_set_double, double, G_TYPE_DOUBLE, g_va
  *
  * Sets a single-valued integer object.
  */
-SET_PROPERTY_FOR_GTYPE (tracker_resource_set_int, int, G_TYPE_INT, g_value_set_int, validate_int);
+SET_PROPERTY_FOR_GTYPE (tracker_resource_set_int, int, G_TYPE_INT, g_value_set_int, validate_int)
 
 /**
  * tracker_resource_set_int64:
@@ -390,7 +395,7 @@ SET_PROPERTY_FOR_GTYPE (tracker_resource_set_int, int, G_TYPE_INT, g_value_set_i
  *
  * Sets a single-valued integer object.
  */
-SET_PROPERTY_FOR_GTYPE (tracker_resource_set_int64, gint64, G_TYPE_INT64, g_value_set_int64, validate_int64);
+SET_PROPERTY_FOR_GTYPE (tracker_resource_set_int64, gint64, G_TYPE_INT64, g_value_set_int64, validate_int64)
 
 /**
  * tracker_resource_set_relation:
@@ -403,7 +408,7 @@ SET_PROPERTY_FOR_GTYPE (tracker_resource_set_int64, gint64, G_TYPE_INT64, g_valu
  * although in this function the URI will depend on the identifier
  * set on @resource.
  */
-SET_PROPERTY_FOR_GTYPE (tracker_resource_set_relation, TrackerResource *, TRACKER_TYPE_RESOURCE, g_value_set_object, validate_pointer);
+SET_PROPERTY_FOR_GTYPE (tracker_resource_set_relation, TrackerResource *, TRACKER_TYPE_RESOURCE, g_value_set_object, validate_pointer)
 
 /**
  * tracker_resource_set_take_relation: (skip)
@@ -416,7 +421,7 @@ SET_PROPERTY_FOR_GTYPE (tracker_resource_set_relation, TrackerResource *, TRACKE
  * although in this function the URI will depend on the identifier
  * set on @resource. This function takes ownership of @resource.
  */
-SET_PROPERTY_FOR_GTYPE (tracker_resource_set_take_relation, TrackerResource *, TRACKER_TYPE_RESOURCE, g_value_take_object, validate_pointer);
+SET_PROPERTY_FOR_GTYPE (tracker_resource_set_take_relation, TrackerResource *, TRACKER_TYPE_RESOURCE, g_value_take_object, validate_pointer)
 
 /**
  * tracker_resource_set_string:
@@ -426,7 +431,7 @@ SET_PROPERTY_FOR_GTYPE (tracker_resource_set_take_relation, TrackerResource *, T
  *
  * Sets a single-valued string object.
  */
-SET_PROPERTY_FOR_GTYPE (tracker_resource_set_string, const char *, G_TYPE_STRING, g_value_set_string, validate_pointer);
+SET_PROPERTY_FOR_GTYPE (tracker_resource_set_string, const char *, G_TYPE_STRING, g_value_set_string, validate_pointer)
 
 /**
  * tracker_resource_set_uri:
@@ -438,7 +443,18 @@ SET_PROPERTY_FOR_GTYPE (tracker_resource_set_string, const char *, G_TYPE_STRING
  * produces similar RDF to tracker_resource_set_relation(), although
  * it requires that the URI is previously known.
  */
-SET_PROPERTY_FOR_GTYPE (tracker_resource_set_uri, const char *, TRACKER_TYPE_URI, g_value_set_string, validate_pointer);
+SET_PROPERTY_FOR_GTYPE (tracker_resource_set_uri, const char *, TRACKER_TYPE_URI, g_value_set_string, validate_pointer)
+
+/**
+ * tracker_resource_set_datetime:
+ * @self: the #TrackerResource
+ * @property_uri: a string identifying the property to modify
+ * @value: the property object
+ *
+ * Sets a single-valued GDateTime as a #TrackerResource
+ * Since: 3.2
+ */
+SET_PROPERTY_FOR_GTYPE (tracker_resource_set_datetime, GDateTime *, G_TYPE_DATE_TIME, g_value_set_boxed, validate_pointer)
 
 /**
  * tracker_resource_add_gvalue:
@@ -499,7 +515,7 @@ tracker_resource_add_gvalue (TrackerResource *self,
 	if (array_holder != existing_value) {
 		g_hash_table_insert (priv->properties, g_strdup (property_uri), array_holder);
 	}
-};
+}
 
 #define ADD_PROPERTY_FOR_GTYPE(name, ctype, gtype, set_function, validate_function)     \
 	void name (TrackerResource *self,                                               \
@@ -552,7 +568,7 @@ tracker_resource_add_gvalue (TrackerResource *self,
 			g_hash_table_insert (priv->properties,                          \
 			                     g_strdup (property_uri), array_holder);    \
 		}                                                                       \
-	};
+	}
 
 /**
  * tracker_resource_add_boolean:
@@ -562,7 +578,7 @@ tracker_resource_add_gvalue (TrackerResource *self,
  *
  * Adds a boolean object to a multi-valued property.
  */
-ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_boolean, gboolean, G_TYPE_BOOLEAN, g_value_set_boolean, validate_boolean);
+ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_boolean, gboolean, G_TYPE_BOOLEAN, g_value_set_boolean, validate_boolean)
 
 /**
  * tracker_resource_add_double:
@@ -572,7 +588,7 @@ ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_boolean, gboolean, G_TYPE_BOOLEAN,
  *
  * Adds a double object to a multi-valued property.
  */
-ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_double, double, G_TYPE_DOUBLE, g_value_set_double, validate_double);
+ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_double, double, G_TYPE_DOUBLE, g_value_set_double, validate_double)
 
 /**
  * tracker_resource_add_int:
@@ -582,7 +598,7 @@ ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_double, double, G_TYPE_DOUBLE, g_va
  *
  * Adds an integer object to a multi-valued property.
  */
-ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_int, int, G_TYPE_INT, g_value_set_int, validate_int);
+ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_int, int, G_TYPE_INT, g_value_set_int, validate_int)
 
 /**
  * tracker_resource_add_int64:
@@ -592,7 +608,7 @@ ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_int, int, G_TYPE_INT, g_value_set_i
  *
  * Adds an integer object to a multi-valued property.
  */
-ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_int64, gint64, G_TYPE_INT64, g_value_set_int64, validate_int64);
+ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_int64, gint64, G_TYPE_INT64, g_value_set_int64, validate_int64)
 
 /**
  * tracker_resource_add_relation:
@@ -605,7 +621,7 @@ ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_int64, gint64, G_TYPE_INT64, g_valu
  * although in this function the URI will depend on the identifier
  * set on @resource.
  */
-ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_relation, TrackerResource *, TRACKER_TYPE_RESOURCE, g_value_set_object, validate_pointer);
+ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_relation, TrackerResource *, TRACKER_TYPE_RESOURCE, g_value_set_object, validate_pointer)
 
 /**
  * tracker_resource_add_take_relation: (skip)
@@ -618,7 +634,7 @@ ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_relation, TrackerResource *, TRACKE
  * although in this function the URI will depend on the identifier
  * set on @resource. This function takes ownership of @resource.
  */
-ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_take_relation, TrackerResource *, TRACKER_TYPE_RESOURCE, g_value_take_object, validate_pointer);
+ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_take_relation, TrackerResource *, TRACKER_TYPE_RESOURCE, g_value_take_object, validate_pointer)
 
 
 /**
@@ -629,7 +645,7 @@ ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_take_relation, TrackerResource *, T
  *
  * Adds a string object to a multi-valued property.
  */
-ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_string, const char *, G_TYPE_STRING, g_value_set_string, validate_pointer);
+ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_string, const char *, G_TYPE_STRING, g_value_set_string, validate_pointer)
 
 /**
  * tracker_resource_add_uri:
@@ -641,8 +657,18 @@ ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_string, const char *, G_TYPE_STRING
  * produces similar RDF to tracker_resource_add_relation(), although
  * it requires that the URI is previously known.
  */
-ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_uri, const char *, TRACKER_TYPE_URI, g_value_set_string, validate_pointer);
+ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_uri, const char *, TRACKER_TYPE_URI, g_value_set_string, validate_pointer)
 
+/**
+ * tracker_resource_add_datetime:
+ * @self: the #TrackerResource
+ * @property_uri: a string identifying the property to modify
+ * @value: the property object
+ *
+ * Adds GDateTime object to the multi-valued property.
+ * Since: 3.2
+ */
+ADD_PROPERTY_FOR_GTYPE (tracker_resource_add_datetime, GDateTime *, G_TYPE_DATE_TIME, g_value_set_boxed, validate_pointer)
 
 /**
  * tracker_resource_get_values:
@@ -674,7 +700,7 @@ GList *tracker_resource_get_values (TrackerResource *self,
 	if (G_VALUE_HOLDS (value, G_TYPE_PTR_ARRAY)) {
 		GList *result = NULL;
 		GPtrArray *array;
-		int i;
+		guint i;
 
 		array = g_value_get_boxed (value);
 
@@ -718,7 +744,7 @@ GList *tracker_resource_get_values (TrackerResource *self,
 		}                                                             \
                                                                               \
 		return get_function (value);                                  \
-	};
+	}
 
 /**
  * tracker_resource_get_first_boolean:
@@ -729,7 +755,7 @@ GList *tracker_resource_get_values (TrackerResource *self,
  *
  * Returns: the first boolean object
  */
-GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_boolean, gboolean, G_TYPE_BOOLEAN, g_value_get_boolean, FALSE);
+GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_boolean, gboolean, G_TYPE_BOOLEAN, g_value_get_boolean, FALSE)
 
 /**
  * tracker_resource_get_first_double:
@@ -740,7 +766,7 @@ GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_boolean, gboolean, G_TYPE_BOO
  *
  * Returns: the first double object
  */
-GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_double, double, G_TYPE_DOUBLE, g_value_get_double, 0.0);
+GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_double, double, G_TYPE_DOUBLE, g_value_get_double, 0.0)
 
 /**
  * tracker_resource_get_first_int:
@@ -751,7 +777,7 @@ GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_double, double, G_TYPE_DOUBLE
  *
  * Returns: the first integer object
  */
-GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_int, int, G_TYPE_INT, g_value_get_int, 0);
+GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_int, int, G_TYPE_INT, g_value_get_int, 0)
 
 /**
  * tracker_resource_get_first_int64:
@@ -762,7 +788,7 @@ GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_int, int, G_TYPE_INT, g_value
  *
  * Returns: the first integer object
  */
-GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_int64, gint64, G_TYPE_INT64, g_value_get_int64, 0);
+GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_int64, gint64, G_TYPE_INT64, g_value_get_int64, 0)
 
 /**
  * tracker_resource_get_first_relation:
@@ -773,7 +799,7 @@ GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_int64, gint64, G_TYPE_INT64,
  *
  * Returns: (transfer none): the first resource object
  */
-GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_relation, TrackerResource *, TRACKER_TYPE_RESOURCE, g_value_get_object, NULL);
+GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_relation, TrackerResource *, TRACKER_TYPE_RESOURCE, g_value_get_object, NULL)
 
 /**
  * tracker_resource_get_first_string:
@@ -784,7 +810,7 @@ GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_relation, TrackerResource *,
  *
  * Returns: the first string object
  */
-GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_string, const char *, G_TYPE_STRING, g_value_get_string, NULL);
+GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_string, const char *, G_TYPE_STRING, g_value_get_string, NULL)
 
 /**
  * tracker_resource_get_first_uri:
@@ -795,7 +821,19 @@ GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_string, const char *, G_TYPE_
  *
  * Returns: the first resource object as an URI.
  */
-GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_uri, const char *, TRACKER_TYPE_URI, g_value_get_string, NULL);
+GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_uri, const char *, TRACKER_TYPE_URI, g_value_get_string, NULL)
+
+/**
+ * tracker_resource_get_first_datetime:
+ * @self: A #TrackerResource
+ * @property_uri: a string identifying the property to look up
+ *
+ * Returns the first resource object previously assigned to a property.
+ *
+ * Returns: the first GDateTime object
+ * Since: 3.2
+ */
+GET_PROPERTY_FOR_GTYPE (tracker_resource_get_first_datetime, GDateTime *, G_TYPE_DATE_TIME, g_value_get_boxed, NULL)
 
 /**
  * tracker_resource_get_identifier:
@@ -909,7 +947,7 @@ tracker_resource_compare (TrackerResource *a,
 	b_priv = GET_PRIVATE (b);
 
 	return strcmp (a_priv->identifier, b_priv->identifier);
-};
+}
 
 /**
  * tracker_resource_get_properties:
@@ -934,6 +972,26 @@ tracker_resource_get_properties (TrackerResource *resource)
 	return g_hash_table_get_keys (priv->properties);
 }
 
+static gchar *
+parse_prefix (const gchar *prefixed_name)
+{
+	const gchar *end, *token_end;
+
+	end = &prefixed_name[strlen(prefixed_name)];
+
+	if (!terminal_PNAME_NS (prefixed_name, end, &token_end))
+		return NULL;
+
+	/* We have read the ':', take a step back */
+	if (token_end && token_end > prefixed_name)
+		token_end--;
+
+	if (*token_end != ':')
+		return NULL;
+
+	return g_strndup (prefixed_name, token_end - prefixed_name);
+}
+
 /* Helper function for serialization code. This allows you to selectively
  * populate 'interned_namespaces' from 'all_namespaces' based on when a
  * particular prefix is actually used. This is quite inefficient compared
@@ -952,7 +1010,7 @@ maybe_intern_prefix_of_compact_uri (TrackerNamespaceManager *all_namespaces,
 	 * we can't really tell if the user has done something dumb like defining a
 	 * "urn" prefix.
 	 */
-	char *prefix = g_uri_parse_scheme (uri);
+	char *prefix = parse_prefix (uri);
 
 	if (prefix == NULL) {
 		g_warning ("Invalid URI or compact URI: %s", uri);
@@ -970,14 +1028,6 @@ maybe_intern_prefix_of_compact_uri (TrackerNamespaceManager *all_namespaces,
 }
 
 
-typedef struct {
-	TrackerNamespaceManager *all_namespaces, *our_namespaces;
-	GString *string;
-	GList *done_list;
-} GenerateTurtleData;
-
-static void generate_turtle (TrackerResource *resource, GenerateTurtleData *data);
-
 static gboolean
 is_blank_node (const char *uri_or_curie_or_blank)
 {
@@ -991,12 +1041,13 @@ is_builtin_class (const gchar             *uri_or_curie,
 	gchar *prefix = NULL;
 	gboolean has_prefix;
 
-	// blank nodes should be processed as nested resource
-	// g_uri_parse_scheme returns NULL for blank nodes, i.e. _:1
+	/* blank nodes should be processed as nested resource
+	 * parse_prefix returns NULL for blank nodes, i.e. _:1
+	 */
 	if (is_blank_node (uri_or_curie))
 		return FALSE;
 
-	prefix = g_uri_parse_scheme (uri_or_curie);
+	prefix = parse_prefix (uri_or_curie);
 
 	if (!prefix)
 		return TRUE;
@@ -1007,50 +1058,6 @@ is_builtin_class (const gchar             *uri_or_curie,
 	return has_prefix;
 }
 
-static void
-generate_nested_turtle_resource (TrackerResource    *resource,
-                                 GenerateTurtleData *data)
-{
-	/* We don't need to produce turtle for builtin classes */
-	if (is_builtin_class (tracker_resource_get_identifier (resource),
-	                      data->all_namespaces))
-		return;
-
-	if (g_list_find (data->done_list, resource) == NULL) {
-		data->done_list = g_list_prepend (data->done_list, resource);
-		generate_turtle (resource, data);
-		g_string_append (data->string, "\n");
-	}
-}
-
-static void
-generate_turtle_resources_foreach (gpointer key,
-                                   gpointer value_ptr,
-                                   gpointer user_data)
-{
-	const GValue *value = value_ptr;
-	GenerateTurtleData *data = user_data;
-	TrackerResource *resource;
-	int i;
-
-	if (G_VALUE_HOLDS (value, TRACKER_TYPE_RESOURCE)) {
-		resource = g_value_get_object (value);
-		generate_nested_turtle_resource (resource, data);
-	} else if (G_VALUE_HOLDS (value, G_TYPE_PTR_ARRAY)) {
-		GPtrArray *array = g_value_get_boxed (value);
-		const GValue *array_value;
-
-		for (i = 0; i < array->len; i++) {
-			array_value = g_ptr_array_index (array, i);
-
-			if (G_VALUE_HOLDS (array_value, TRACKER_TYPE_RESOURCE)) {
-				resource = g_value_get_object (array_value);
-				generate_nested_turtle_resource (resource, data);
-			}
-		}
-	}
-}
-
 static void
 generate_turtle_uri_value (const char              *uri_or_curie_or_blank,
                            GString                 *string,
@@ -1069,7 +1076,7 @@ generate_turtle_uri_value (const char              *uri_or_curie_or_blank,
 	if (is_blank_node (uri_or_curie_or_blank)) {
 		g_string_append (string, uri_or_curie_or_blank);
 	} else {
-		char *prefix = g_uri_parse_scheme (uri_or_curie_or_blank);
+		char *prefix = parse_prefix (uri_or_curie_or_blank);
 
 		if (prefix && tracker_namespace_manager_has_prefix (all_namespaces, prefix)) {
 			/* It's a compact URI and we know the prefix */
@@ -1154,7 +1161,7 @@ generate_turtle_property (const char              *property,
 
 	g_string_append (string, " ");
 	if (G_VALUE_HOLDS (value, G_TYPE_PTR_ARRAY)) {
-		int i;
+		guint i;
 		GPtrArray *array = g_value_get_boxed (value);
 		if (array->len > 0) {
 			generate_turtle_value (g_ptr_array_index (array, 0),
@@ -1174,44 +1181,11 @@ generate_turtle_property (const char              *property,
 	}
 }
 
-static void
-generate_turtle (TrackerResource    *resource,
-                 GenerateTurtleData *data)
-{
-	TrackerResourcePrivate *priv = GET_PRIVATE (resource);
-	GHashTableIter iter;
-	const char *property;
-	const GValue *value;
-
-	/* First we recurse to any relations that aren't already in the done list */
-	g_hash_table_foreach (priv->properties, generate_turtle_resources_foreach, data);
-
-	generate_turtle_uri_value (tracker_resource_get_identifier(resource),
-	        data->string, data->all_namespaces, data->our_namespaces);
-	g_string_append (data->string, " ");
-
-	g_hash_table_iter_init (&iter, priv->properties);
-	if (g_hash_table_iter_next (&iter, (gpointer *)&property, (gpointer *)&value)) {
-		while (TRUE) {
-			generate_turtle_property (property, value, data->string, data->all_namespaces, data->our_namespaces);
-
-			maybe_intern_prefix_of_compact_uri (data->all_namespaces, data->our_namespaces, property);
-
-			if (g_hash_table_iter_next (&iter, (gpointer *)&property, (gpointer *)&value)) {
-				g_string_append (data->string, " ;\n  ");
-			} else {
-				g_string_append (data->string, " .\n");
-				break;
-			}
-		}
-	}
-}
-
 /**
  * tracker_resource_print_turtle:
  * @self: a #TrackerResource
  * @namespaces: (allow-none): a set of prefixed URLs, or %NULL to use the
- *     default set
+ *     Nepomuk set
  *
  * Serialize all the information in @resource as a Turtle document.
  *
@@ -1219,51 +1193,24 @@ generate_turtle (TrackerResource    *resource,
  * <https://www.w3.org/TR/2014/REC-turtle-20140225/>
  *
  * The @namespaces object is used to expand any compact URI values. In most
- * cases you should pass %NULL, which means the set of namespaces returned by
- * tracker_namespace_manager_get_default() will be used. This defines the
- * usual prefixes for all of the ontologies that Tracker ships with by default.
+ * cases you should pass the one returned by tracker_sparql_connection_get_namespace_manager()
+ * from the connection that is the intended recipient of this data.
  *
  * Returns: a newly-allocated string
+ *
+ * Deprecated: 3.4: Use tracker_resource_print_rdf() instead.
  */
 char *
 tracker_resource_print_turtle (TrackerResource         *self,
                                TrackerNamespaceManager *namespaces)
 {
-	TrackerResourcePrivate *priv;
-	GenerateTurtleData context;
-	char *prefixes;
-
 	g_return_val_if_fail (TRACKER_IS_RESOURCE (self), "");
 
-	priv = GET_PRIVATE (self);
-
 	if (namespaces == NULL) {
 		namespaces = tracker_namespace_manager_get_default ();
 	}
 
-	if (g_hash_table_size (priv->properties) == 0) {
-		return g_strdup("");
-	}
-
-	context.all_namespaces = namespaces;
-	context.our_namespaces = tracker_namespace_manager_new ();
-	context.string = g_string_new ("");
-	context.done_list = g_list_prepend (NULL, self);
-
-	maybe_intern_prefix_of_compact_uri (context.all_namespaces, context.our_namespaces, tracker_resource_get_identifier(self));
-
-	generate_turtle (self, &context);
-
-	prefixes = tracker_namespace_manager_print_turtle (context.our_namespaces);
-	g_string_prepend (context.string, "\n");
-	g_string_prepend (context.string, prefixes);
-
-	g_object_unref (context.our_namespaces);
-	g_free (prefixes);
-
-	g_list_free (context.done_list);
-
-	return g_string_free (context.string, FALSE);
+	return tracker_resource_print_rdf (self, namespaces, TRACKER_RDF_FORMAT_TURTLE, NULL);
 }
 
 typedef struct {
@@ -1283,7 +1230,7 @@ generate_sparql_relation_deletes_foreach (gpointer key,
 {
 	const GValue *value = value_ptr;
 	GenerateSparqlData *data = user_data;
-	int i;
+	guint i;
 
 	if (G_VALUE_HOLDS (value, TRACKER_TYPE_RESOURCE)) {
 		TrackerResource *relation = g_value_get_object (value);
@@ -1325,7 +1272,7 @@ generate_sparql_relation_inserts_foreach (gpointer key,
 		GPtrArray *array = g_value_get_boxed (value);
 		const GValue *array_value;
 		TrackerResource *relation;
-		gint i;
+		guint i;
 
 		for (i = 0; i < array->len; i++) {
 			array_value = g_ptr_array_index (array, i);
@@ -1466,7 +1413,7 @@ generate_sparql_insert_pattern (TrackerResource    *resource,
  * tracker_resource_print_sparql_update:
  * @self: a #TrackerResource
  * @namespaces: (allow-none): a set of prefixed URLs, or %NULL to use the
- *     default set
+ *     Nepomuk set
  * @graph_id: (allow-none): the URN of the graph the data should be added to,
  *     or %NULL
  *
@@ -1474,9 +1421,8 @@ generate_sparql_insert_pattern (TrackerResource    *resource,
  * stored in @resource.
  *
  * The @namespaces object is used to expand any compact URI values. In most
- * cases you should pass %NULL, which means the set of namespaces returned by
- * tracker_namespace_manager_get_default() will be used. This defines the
- * usual prefixes for all of the ontologies that Tracker ships with by default.
+ * cases you should pass the one returned by tracker_sparql_connection_get_namespace_manager()
+ * from the connection that is the intended recipient of this data.
  *
  * Returns: a newly-allocated string containing a SPARQL update command.
  */
@@ -1566,7 +1512,7 @@ tracker_resource_generate_jsonld (TrackerResource    *self,
 	}
 
 	g_hash_table_foreach (priv->properties, generate_jsonld_foreach, data);
-};
+}
 
 static void
 generate_jsonld_value (const GValue       *value,
@@ -1654,7 +1600,7 @@ generate_jsonld_namespace_mapping_foreach (gpointer key,
  * tracker_resource_print_jsonld:
  * @self: a #TrackerResource
  * @namespaces: (allow-none): a set of prefixed URLs, or %NULL to use the
- *     default set
+ *     Nepomuk set
  *
  * Serialize all the information in @resource as a JSON-LD document.
  *
@@ -1662,9 +1608,8 @@ generate_jsonld_namespace_mapping_foreach (gpointer key,
  * serialization format.
  *
  * The @namespaces object is used to expand any compact URI values. In most
- * cases you should pass %NULL, which means the set of namespaces returned by
- * tracker_namespace_manager_get_default() will be used. This defines the
- * usual prefixes for all of the ontologies that Tracker ships with by default.
+ * cases you should pass the one returned by tracker_sparql_connection_get_namespace_manager()
+ * from the connection that is the intended recipient of this data.
  *
  * Returns: a newly-allocated string containing JSON-LD data.
  */
@@ -1715,6 +1660,87 @@ tracker_resource_print_jsonld (TrackerResource         *self,
 	return result;
 }
 
+static TrackerSerializerFormat
+convert_format (TrackerRdfFormat format)
+{
+	switch (format) {
+	case TRACKER_RDF_FORMAT_TURTLE:
+		return TRACKER_SERIALIZER_FORMAT_TTL;
+	case TRACKER_RDF_FORMAT_TRIG:
+		return TRACKER_SERIALIZER_FORMAT_TRIG;
+	case TRACKER_N_RDF_FORMATS:
+		g_assert_not_reached ();
+	}
+
+	return -1;
+}
+
+/**
+ * tracker_resource_print_rdf:
+ * @self: a #TrackerResource
+ * @namespaces: a set of prefixed URLs
+ * @format: RDF format of the printed string
+ * @graph: target graph of the resource RDF, or %NULL for the default graph
+ *
+ * Serialize all the information in @resource into the selected RDF format.
+ *
+ * The @namespaces object is used to expand any compact URI values. In most
+ * cases you should pass the one returned by tracker_sparql_connection_get_namespace_manager()
+ * from the connection that is the intended recipient of this data.
+ *
+ * Returns: a newly-allocated string containing RDF data in the requested format.
+ *
+ * Since: 3.4
+ **/
+char *
+tracker_resource_print_rdf (TrackerResource         *self,
+                            TrackerNamespaceManager *namespaces,
+                            TrackerRdfFormat         format,
+                            const gchar             *graph)
+{
+	TrackerSparqlCursor *deserializer;
+	GInputStream *serializer;
+	GString *str;
+
+	g_return_val_if_fail (TRACKER_IS_RESOURCE (self), NULL);
+	g_return_val_if_fail (TRACKER_IS_NAMESPACE_MANAGER (namespaces), NULL);
+	g_return_val_if_fail (format < TRACKER_N_RDF_FORMATS, NULL);
+
+#define BUF_SIZE 4096
+	deserializer = tracker_deserializer_resource_new (self, namespaces, graph);
+	serializer = tracker_serializer_new (TRACKER_SPARQL_CURSOR (deserializer),
+	                                     namespaces,
+	                                     convert_format (format));
+	g_object_unref (deserializer);
+
+	str = g_string_new (NULL);
+
+	while (TRUE) {
+		GBytes *bytes;
+
+		bytes = g_input_stream_read_bytes (serializer, BUF_SIZE, NULL, NULL);
+		if (!bytes) {
+			g_string_free (str, TRUE);
+			return NULL;
+		}
+
+		if (g_bytes_get_size (bytes) == 0) {
+			g_bytes_unref (bytes);
+			break;
+		}
+
+		g_string_append_len (str,
+		                     g_bytes_get_data (bytes, NULL),
+		                     g_bytes_get_size (bytes));
+		g_bytes_unref (bytes);
+	}
+#undef BUF_SIZE
+
+	g_object_unref (serializer);
+
+	return g_string_free (str, FALSE);
+}
+
 static GVariant *
 tracker_serialize_single_value (TrackerResource         *resource,
                                 const GValue            *value)
@@ -1961,3 +1987,50 @@ tracker_resource_get_property_overwrite (TrackerResource *resource,
 
 	return g_hash_table_contains (priv->overwrite, property_uri);
 }
+
+void
+tracker_resource_iterator_init (TrackerResourceIterator *iter,
+                                TrackerResource         *resource)
+{
+	TrackerResourcePrivate *priv = GET_PRIVATE (resource);
+
+	bzero (iter, sizeof (TrackerResourceIterator));
+	g_hash_table_iter_init (&iter->prop_iter, priv->properties);
+}
+
+gboolean
+tracker_resource_iterator_next (TrackerResourceIterator  *iter,
+                                const gchar             **property,
+                                const GValue            **value)
+{
+	gpointer key, val;
+
+	if (iter->cur_values && iter->cur_prop) {
+		iter->idx++;
+
+		if (iter->idx < iter->cur_values->len) {
+			*property = iter->cur_prop;
+			*value = g_ptr_array_index (iter->cur_values, iter->idx);
+			return TRUE;
+		} else {
+			iter->cur_values = NULL;
+			iter->cur_prop = NULL;
+		}
+	}
+
+	if (!g_hash_table_iter_next (&iter->prop_iter, &key, &val))
+		return FALSE;
+
+	if (G_VALUE_HOLDS (val, G_TYPE_PTR_ARRAY)) {
+		iter->cur_prop = key;
+		iter->cur_values = g_value_get_boxed (val);
+		iter->idx = 0;
+		*property = iter->cur_prop;
+		*value = g_ptr_array_index (iter->cur_values, iter->idx);
+		return TRUE;
+	}
+
+	*property = key;
+	*value = val;
+	return TRUE;
+}
